#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "utils.h"
#include "alifold.h"
#include "structs.h"

extern struct seqStruct *seqLst;
extern struct conStruct constraint;
extern struct ALNdata aln;
extern struct input_options options;
extern int fold_constrained;
extern int alifold_count;
extern int number_of_seqs;

void usage(void);
void alignPairwise(int n, int c);
void shrink_to_seq(int a);
void expand_to_aln(int i, int j);
/* void strictConsensus(int n, int c); */ /* unused! */
char *pairwiseConsensus(int i, int j);
long align(int a, int b);
void freePairwiseAlnStruct(void);
void severe_error(int a);

/*-----------------*/

void alignPairwise(int n, int c) {

  int i=0, j=0, a=0, x=0, y=0, m, p;
  int *count, wait=0;   int aa;
  long z, q, q2;
  char **sequences=NULL, *consStruc=NULL, *copy_consStruc=NULL, *consStruc0=NULL, *consStruc1=NULL;
  float mfe;
  
  count = (int*)space(number_of_seqs * sizeof(int));
  
  for(i=0; i < number_of_seqs; i++){
    if(i == c) continue;    /* ignore constraint */
    count[i] = 0;
    seqLst[i].weakConstr = NULL;
    seqLst[i].weakConstr = (char*)space((strlen(constraint.weakened)+1)*sizeof(char));
    strcpy(seqLst[i].weakConstr, constraint.weakened);
    shrink_to_seq(i);
  }
  
  for(i=0; i<n-2; i++) {
    for(j = i+1; j<n-1; j++) {
      wait++;
    }
  }
  fprintf(stdout, "  (%i alignment foldings will be done.)\n", wait);
  
  for(i=0; i<n-1; i++) {
    for(j = i+1; j<n; j++) {
      if(i == c || j == c) continue;    /* ignore constraint */
      
      aln.lines = NULL;
      aln.lines = (char**)space(3*sizeof(char*));
      aln.lines[0] = NULL;
      aln.lines[1] = NULL;
      aln.lines[2] = NULL;
      
      if(options.kurzschluss == 0){   /* norm incl. my own aln. */
	fprintf(stdout, "  Alignment %i of %s and %s ..\n", 
		alifold_count + 1, seqLst[i].name, seqLst[j].name);
	z = align(i, j);
	aln.lines[0] = (char*) space(z*sizeof(char));
	aln.lines[1] = (char*) space(z*sizeof(char));
	
	q=0;
	while(aln.alignment[0][q].nucl){
	  aln.lines[0][q] = aln.alignment[0][q].nucl;
	  q++;
	}
	aln.lines[0][q] = '\0';
	
	q=0;
	while(aln.alignment[1][q].nucl){
	  aln.lines[1][q] = aln.alignment[1][q].nucl;
	  q++;
	}
	aln.lines[1][q] = '\0';
      }
      
      else if(options.kurzschluss == 1){      /* KURZSCHLUSS */
	z = strlen(seqLst[i].alnSequence) + 1;
	aln.lines[0] = (char*) space(z*sizeof(char));
	aln.lines[1] = (char*) space(z*sizeof(char));
	strcpy(aln.lines[0], seqLst[i].alnSequence);
	strcpy(aln.lines[1], seqLst[j].alnSequence);
	
	if((strlen(aln.lines[0]) != strlen(aln.lines[1])) || 
	   strlen(aln.lines[0]) != strlen(constraint.weakened)){
	  severe_error(5);
	}
	
	q2=0;
	for(q=0; q < strlen(aln.lines[0]); q++){
	  if(isalpha(aln.lines[0][q]) || isalpha(aln.lines[1][q])){
	    aln.lines[0][q2] = aln.lines[0][q];
	    aln.lines[1][q2] = aln.lines[1][q];
	    q2++;
	  }
	  else continue;
	}
	aln.lines[0][q2] = '\0';
	aln.lines[1][q2] = '\0';
      }
      
      expand_to_aln(i, j);
      
      consStruc = pairwiseConsensus(i,j);
      
      if(strlen(aln.lines[0]) != strlen(seqLst[i].weakConstr2))  severe_error(23);

      free(seqLst[i].weakConstr2);
      free(seqLst[j].weakConstr2);
      
      sequences = NULL;
      sequences = (char**) space(3*sizeof(char*));
      sequences[0] = NULL;
      sequences[1] = NULL;
      sequences[0] = (char*) space((strlen(aln.lines[0])+1)*sizeof(char));
      sequences[1] = (char*) space((strlen(aln.lines[1])+1)*sizeof(char));
      sequences[2] = NULL;
      strcpy(sequences[0], aln.lines[0]);
      strcpy(sequences[1], aln.lines[1]);
      
      if(options.kurzschluss)  fprintf(stdout,"    Constraint directed folding of %s and %s ..\n", seqLst[i].name, seqLst[j].name);
      else   fprintf(stdout,"    Constraint directed folding of alignment ..\n");
      
      copy_consStruc = (char*)space((strlen(consStruc)+1)*sizeof(char));
      strcpy(copy_consStruc, consStruc);
      
      mfe = alifold(sequences, consStruc);
      
      for(aa = 0; aa < strlen(consStruc); aa++){
	if(copy_consStruc[aa] == 'x'){
	  if(consStruc[aa] == '.')    consStruc[aa] = 'x';
	  else severe_error(122);
	}
      }	
      free(copy_consStruc);
      copy_consStruc = NULL;
      
      alifold_count++;
      
      free_alifold_arrays();
      
      fprintf(stdout, "    . done. Minimum free energy is %f.\n", mfe);
      
      /* shrink constraint to sequence length without gaps. */
      
      consStruc0 = NULL; 
      consStruc1 = NULL;
      consStruc0 = (char*)space((strlen(consStruc)+1)*sizeof(char));
      consStruc1 = (char*)space((strlen(consStruc)+1)*sizeof(char));
      strcpy(consStruc0, consStruc);
      strcpy(consStruc1, consStruc);
      
      if(strlen(consStruc) != strlen(sequences[0]))   severe_error(88);
      
      for(a=0; a<strlen(sequences[0]); a++) {
	if(sequences[0][a] != '-') continue;
	else {
	  switch(consStruc0[a]){
	  case 'x':
	    break;
	  case '.':
	    break;
	  case '(':
	    m=1;
	    p=a+1;
	    while(m){
	      if(consStruc0[p] == '(') m++;
	      if(consStruc0[p] == ')') m--;
	      p++; 
	    }
	    consStruc0[p-1] = '.';
	    consStruc0[a] = '.';
	    break;
	  case ')':
	    m=1;
	    p=a-1;
	    while(m){
	      if(consStruc0[p] == '(') m--;
	      if(consStruc0[p] == ')') m++;
	      p--; 
	    }
	    consStruc0[p+1] = '.';
	    consStruc0[a] = '.';
	    break;
	  default: severe_error(33); /* strange signs in structure. */
	  }
	}
      }
      
      for(a=0; a<strlen(sequences[1]); a++) {
	if(sequences[1][a] != '-') continue;
	else {
	  switch(consStruc1[a]){
	  case 'x':
	    break;
	  case '.':
	    break;
	  case '(':
	    m=1;
	    p=a+1;
	    while(m){
	      if(consStruc1[p] == '(') m++;
	      if(consStruc1[p] == ')') m--;
	      p++; 
	    }
	    consStruc1[p-1] = '.';
	    consStruc1[a] = '.';
	    break;
	  case ')':
	    m=1;
	    p=a-1;
	    while(m){
	      if(consStruc1[p] == '(') m--;
	      if(consStruc1[p] == ')') m++;
	      p--; 
	    }
	    consStruc1[p+1] = '.';
	    consStruc1[a] = '.';
	    break;
	  default: severe_error(33); /* strange signs in structure. */
	  }
	}
      } 
      
      x = 0; y = 0;
      for(a=0; a<strlen(sequences[0]); a++) {
	if(sequences[0][a] != '-') {
	  seqLst[i].consStrucLst[count[i]][x] = consStruc0[a];
	  x++;
	}
	if(sequences[1][a] != '-') {
	  seqLst[j].consStrucLst[count[j]][y] = consStruc1[a];
	  y++;
	}
      }
      seqLst[i].consStrucLst[count[i]][x] = '\0';
      seqLst[j].consStrucLst[count[j]][y] = '\0';
      
      if((seqLst[i].length != x) || (seqLst[j].length != y)) severe_error(12); 
      
      count[i]++;
      count[j]++;
      
      free(sequences[0]);
      free(sequences[1]);
      free(sequences);
      free(consStruc0);
      free(consStruc1);
      free(consStruc);
      
      if(options.kurzschluss == 1){  /* KURZSCHLUSS */
	free(aln.lines[0]);
	free(aln.lines[1]);
	free(aln.lines);
      }
      else freePairwiseAlnStruct();
    }
  }
  
  j=0;
  for(i=0; i<n; i++) {
    if(i == c) continue;   
    while(seqLst[i].consStrucLst[j]){
      j++;
    }	
    if(count[i] != j) severe_error(15);
  }
  for(i=0; i < number_of_seqs; i++){
    if(i == c) continue;    /* ignore constraint */
    free(seqLst[i].weakConstr);
  }
  fprintf(stdout, ". done.\n");
}

/*----------------*/

void shrink_to_seq(int a){
  
  int i,b, k=0, m=0, p;    
  char *tmp = NULL, *control_seq = NULL;
  
  /* resolve problem of missing pairing bases within end gaps. */
  
  i=0;
  
  while(seqLst[a].alnSequence[i] == '-'){
    seqLst[a].weakConstr[i] = '.';
    i++;
  }
  i = strlen(seqLst[a].alnSequence) - 1;
  while(seqLst[a].alnSequence[i] == '-'){
    seqLst[a].weakConstr[i] = '.';
    i--;
  }
 
  for(i=0; i<strlen(seqLst[a].alnSequence); i++){
    switch(seqLst[a].weakConstr[i]){
    case '.':
      break;
    case '(':
      m++;
      break; 
    case ')':
      m--;
      if(m < 0){
	seqLst[a].weakConstr[i] = 'x';
	m = 0;
      }
      break; 
    default: break;
    }
  }
  
  m=0;
  
  for(i = strlen(seqLst[a].alnSequence) - 1; i >= 0; i--){
    switch(seqLst[a].weakConstr[i]){
    case '.':
      break;
    case ')':
      m++;
      break; 
    case '(':
      m--;
      if(m < 0){
	seqLst[a].weakConstr[i] = 'x';
	m = 0;
      }
      break; 
    default: break;
    }
  }
  
/* start shrinking itself */
  
  
  b = strlen(seqLst[a].alnSequence);
  
  tmp = (char*)space((b+1)*sizeof(char));
  control_seq = (char*)space((b+1)*sizeof(char));
  
  for(i=0; i<strlen(seqLst[a].alnSequence); i++){
    if(seqLst[a].alnSequence[i] != '-'){
      
      tmp[k] = seqLst[a].weakConstr[i];
      control_seq[k] = seqLst[a].alnSequence[i];
      k++;
    }
    else {
      switch(seqLst[a].weakConstr[i]){
      case '.':
	break;
      case '(':
	m=1;
	p=i+1;
	while(m){
	  if(seqLst[a].weakConstr[p] == '(') m++;
	  if(seqLst[a].weakConstr[p] == ')') m--;
	  p++; 
	}
	seqLst[a].weakConstr[p-1] = '.';
	break;
      case ')':
	m=1;
	p=k-1;
	while(m){
	  if(tmp[p] == '(') m--;
	  if(tmp[p] == ')') m++;
	  p--; 
	}
	tmp[p+1] = '.';
	break;
      case 'x':
	break;
      default: severe_error(134);
      }
    }     
  }
  strcat(tmp, "\0");
  strcat(control_seq, "\0");

  if(strlen(tmp) != strlen(control_seq))  severe_error(90);

  strcpy(seqLst[a].weakConstr, tmp);
  free(tmp);
  free(control_seq);

}

/*----------------*/
  
void expand_to_aln(i, j){

  int m, k=0;

  seqLst[i].weakConstr2 = NULL;
  seqLst[j].weakConstr2 = NULL;
  seqLst[i].weakConstr2 = (char*)space((strlen(aln.lines[0])+1)*sizeof(char));
  seqLst[j].weakConstr2 = (char*)space((strlen(aln.lines[1])+1)*sizeof(char));

  if(strlen(aln.lines[0]) != strlen(aln.lines[1])) severe_error(21);
  
  for(m=0; m < strlen(aln.lines[0]); m++){
    if(aln.lines[0][m] == '-') {
      seqLst[i].weakConstr2[m] = '.';
    }
    else {
      seqLst[i].weakConstr2[m] = seqLst[i].weakConstr[k];
      k++;
    }
  }
  /* strcat(seqLst[i].weakConstr2,"\0"); */
  seqLst[i].weakConstr2[strlen(aln.lines[0])] = '\0';

  if(k != seqLst[i].length)  severe_error(7);
  
  k=0;
  for(m = 0; m < strlen(aln.lines[1]); m++){
    if(aln.lines[1][m] == '-') {
      seqLst[j].weakConstr2[m] = '.';
    }
    else {
      seqLst[j].weakConstr2[m] = seqLst[j].weakConstr[k];
      k++;
    }
  }
  /*  strcat(seqLst[j].weakConstr2, "\0"); */
  seqLst[j].weakConstr2[strlen(aln.lines[1])] = '\0';
  
  if(k != seqLst[j].length) severe_error(8);
}


/*----------------*/

char *pairwiseConsensus(int i, int j) {

  int qq, qqq, a;
  char *pairCons = NULL;

	pairCons = (char*)space((strlen(aln.lines[0])+1)*sizeof(char));
	strcpy(pairCons, seqLst[i].weakConstr2);
	
	while(strcmp(pairCons, seqLst[j].weakConstr2)) {
	for(a=0; a < strlen(pairCons); a++){
		if(pairCons[a] == seqLst[j].weakConstr2[a]) continue;
		else {
			if(pairCons[a] == 'x'){
				if(seqLst[j].weakConstr2[a] == '.'){
					seqLst[j].weakConstr2[a] = 'x';
					continue;
				}
				if(seqLst[j].weakConstr2[a] == '('){
					qqq = 1;
					qq = a + 1;
					while(qqq){
						if(seqLst[j].weakConstr2[qq] == '(') qqq++;
						if(seqLst[j].weakConstr2[qq] == ')') qqq--;
						qq++;
					}
					seqLst[j].weakConstr2[a] = 'x';
					seqLst[j].weakConstr2[qq-1] = 'x';
					continue;
					/* if(pairCons[qq-1] == '.')  pairCons[qq-1] = 'x';
					if(pairCons[qq-1] == '('){
						ccc = 1;
						cc = qq;
						while(ccc){
							if(pairCons[cc] == '(') ccc++;
							if(pairCons[cc] == ')') ccc--;
							cc++;
						}
						pairCons[cc-1] = 'x';
						pairCons[qq-1] = 'x';
					}
					if(pairCons[qq-1] == ')'){
						ccc = 1;
						cc = qq-2;
						while(ccc){
							if(pairCons[cc] == '(') ccc--;
							if(pairCons[cc] == ')') ccc++;
							cc--;
						}
						pairCons[cc+1] = 'x';
						pairCons[qq-1] = 'x';
					}
					seqLst[j].weakConstr2[a] = 'x';
					seqLst[j].weakConstr2[qq-1] = 'x';
					continue; */
				}
				if(seqLst[j].weakConstr2[a] == ')'){
					qqq = 1;
					qq = a - 1;
					while(qqq){
						if(seqLst[j].weakConstr2[qq] == '(') qqq--;
						if(seqLst[j].weakConstr2[qq] == ')') qqq++;
						qq--;
					}
					seqLst[j].weakConstr2[a] = 'x';
					seqLst[j].weakConstr2[qq+1] = 'x';
					continue;
					/* if(pairCons[qq+1] == '.')   pairCons[qq+1] = 'x';
					if(pairCons[qq+1] == '('){
						ccc = 1;
						cc = qq+2;
						while(ccc){
							if(pairCons[cc] == '(') ccc++;
							if(pairCons[cc] == ')') ccc--;
							cc++;
						}
						pairCons[cc-1] = 'x';
						pairCons[qq+1] = 'x';
					}
					if(pairCons[qq+1] == ')'){
						ccc = 1;
						cc = qq;
						while(ccc){
							if(pairCons[cc] == '(') ccc--;
							if(pairCons[cc] == ')') ccc++;
							cc--;
						}
						pairCons[cc+1] = 'x';
						pairCons[qq+1] = 'x';
					}
					seqLst[j].weakConstr2[a] = 'x';
					seqLst[j].weakConstr2[qq+1] = 'x';
					continue; */
				}
				severe_error(120); /* strange sign in structure. */
			}
			if(pairCons[a] == '.'){
				if(seqLst[j].weakConstr2[a] == 'x'){
					pairCons[a] = 'x';
					continue;
				}
				if(seqLst[j].weakConstr2[a] == ')'){
					qqq = 1;
					qq = a - 1;
					while(qqq){
						if(seqLst[j].weakConstr2[qq] == '(') qqq--;
						if(seqLst[j].weakConstr2[qq] == ')') qqq++;
						qq--;
					}
					seqLst[j].weakConstr2[qq+1] = '.';
					seqLst[j].weakConstr2[a] = '.';
					/* if(pairCons[qq+1] == 'x'){
						seqLst[j].weakConstr2[a] = 'x';	
						seqLst[j].weakConstr2[qq-1] = 'x';
					}
					if(pairCons[qq+1] == '.'){
						seqLst[j].weakConstr2[a] = '.';	
						seqLst[j].weakConstr2[qq-1] = '.';
					} */
					continue;
				}
				if(seqLst[j].weakConstr2[a] == '('){
					qqq = 1;
					qq = a + 1;
					while(qqq){
						if(seqLst[j].weakConstr2[qq] == '(') qqq++;
						if(seqLst[j].weakConstr2[qq] == ')') qqq--;
						qq++;
					}
					/* pairCons[qq-1] = '.'; */
					seqLst[j].weakConstr2[a] = '.';
					seqLst[j].weakConstr2[qq-1] = '.';
					continue;
				}
				severe_error(121); /* strange sign in structure. */
			}
			if(pairCons[a] == '('){
				if(seqLst[j].weakConstr2[a] == 'x'){
					qqq = 1;
					qq = a + 1;
					while(qqq){
						if(pairCons[qq] == '(') qqq++;
						if(pairCons[qq] == ')') qqq--;
						qq++;
					}
					pairCons[qq-1] = 'x';
					pairCons[a] = 'x';
					continue;
				}
				if(seqLst[j].weakConstr2[a] == '.'){
					qqq = 1;
					qq = a + 1;
					while(qqq){
						if(pairCons[qq] == '(') qqq++;
						if(pairCons[qq] == ')') qqq--;
						qq++;
					}
					pairCons[qq-1] = '.';
					pairCons[a] = '.';
					continue;
				}
				if(seqLst[j].weakConstr2[a] == ')'){
					qqq = 1;
					qq = a + 1;
					while(qqq){
						if(pairCons[qq] == '(') qqq++;
						if(pairCons[qq] == ')') qqq--;
						qq++;
					}
					pairCons[qq-1] = '.';
					pairCons[a] = '.';
					
					qqq = 1;
					qq = a - 1;
					while(qqq){
						if(seqLst[j].weakConstr2[qq] == '(') qqq--;
						if(seqLst[j].weakConstr2[qq] == ')') qqq++;
						qq--;
					}
					seqLst[j].weakConstr2[qq+1] = '.';
					seqLst[j].weakConstr2[a] = '.';
					/* if(pairCons[qq+1] != 'x')  seqLst[j].weakConstr2[qq+1] = '.';
					else seqLst[j].weakConstr2[qq+1] = 'x'; */
					continue;
				}
				severe_error(122); /* strange sign in structure. */
			}
			if(pairCons[a] == ')'){
				if(seqLst[j].weakConstr2[a] == 'x'){
					qqq = 1;
					qq = a - 1;
					while(qqq){
						if(pairCons[qq] == '(') qqq--;
						if(pairCons[qq] == ')') qqq++;
						qq--;
					}
					pairCons[qq+1] = 'x';
					pairCons[a] = 'x';
					continue;
				}
				if(seqLst[j].weakConstr2[a] == '.'){
					qqq = 1;
					qq = a - 1;
					while(qqq){
						if(pairCons[qq] == '(') qqq--;
						if(pairCons[qq] == ')') qqq++;
						qq--;
					}
					pairCons[qq+1] = '.';
					pairCons[a] = '.';
					continue;
				}
				if(seqLst[j].weakConstr2[a] == '('){
					qqq = 1;
					qq = a - 1;
					while(qqq){
						if(pairCons[qq] == '(') qqq--;
						if(pairCons[qq] == ')') qqq++;
						qq--;
					}
					pairCons[qq+1] = '.';
					pairCons[a] = '.';
					
					qqq = 1;
					qq = a + 1;
					while(qqq){
						if(seqLst[j].weakConstr2[qq] == '(') qqq++;
						if(seqLst[j].weakConstr2[qq] == ')') qqq--;
						qq++;
					}
					seqLst[j].weakConstr2[qq-1] = '.';
					seqLst[j].weakConstr2[a] = '.';
					continue;
				}
				severe_error(123); /* strange sign in structure. */
			}
			severe_error(124); /* strange sign in structure. */
		}
	}
	}
	return(pairCons);
}
	
/*----------------*/

void freePairwiseAlnStruct(void) {
  
  free(aln.alignment[0]);
  free(aln.alignment[1]);
  free(aln.alignment); 
  
  free(aln.lines[0]);
  free(aln.lines[1]);
  free(aln.lines);
  
  /* free(aln.seqname[0]); 
     free(aln.seqname[1]); 
     free(aln.seqname);  */
}

