#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "utils.h"
#include "structs.h"

int readConstraint();
int checkStructure();
int identifyConstraint();
void weakenConstraint(int n, int c);
int scoreColPair(char *c1, char *c2, int n);
void printConstraint();
void freeConstraint();
void file_error(char *file);

extern struct conStruct constraint;
extern struct input_options options;
extern char **seqs;
extern char **names;

/*--------------------------------------------------------------------------*/

int readConstraint() {

    char *line = NULL;
    int valid = 0, i, ch;
    FILE *c_handle = NULL, *out_handle = NULL;

    constraint.name = NULL;
    constraint.sequence = NULL;
    constraint.structure = NULL;
    constraint.expStructure = NULL;
    constraint.weakened = NULL;

    c_handle = fopen(options.constraint_filename, "r");
	out_handle = fopen("tmp.tmp", "w");
    if (c_handle == NULL) {
        file_error(options.constraint_filename);
    }
	while((ch = fgetc (c_handle)) != EOF)	if(ch != '\015' && ch != '\032')   putc(ch, out_handle); 
	fclose(c_handle);
	fclose(out_handle);
	c_handle = fopen("tmp.tmp", "r");
  
    fprintf(stdout, "Reading constraint ..\n");
    line = get_line(c_handle);
    constraint.name = (char*)space((strlen(line)+1)*sizeof(char));
    sscanf(line, "%s", constraint.name);

    if (constraint.name[0] == '>'){
        for (i=1; i<strlen(constraint.name); i++){
            constraint.name[i-1] = constraint.name[i];
        }
        constraint.name[i-1] = '\0';
    }

    strcat(constraint.name, "\0");
    free(line);

    line = get_line(c_handle);
    constraint.sequence = (char*)space((strlen(line)+1)*sizeof(char));
    strcpy(constraint.sequence, line);
    strcat(constraint.sequence, "\0");
    i=0;
    while (constraint.sequence[i] != '\0') {
        constraint.sequence[i] = toupper(constraint.sequence[i]);
        if (constraint.sequence[i] == 'T') constraint.sequence[i] = 'U';
        if (constraint.sequence[i] == '-') {
            fprintf(stderr, "\nError: constraint must not contain gaps. Aborting.\n\n");
            exit(17);
        }
        i++;
    }
    free(line);

    line = get_line(c_handle);
    constraint.structure = (char*)space((strlen(line)+1)*sizeof(char));
    strcpy(constraint.structure, line);
    strcat(constraint.structure, "\0");
    free(line);

    for (i=0; i < strlen(constraint.structure); i++) {
        if ((constraint.structure[i] == '{') || (constraint.structure[i] == '}')){
            constraint.structure[i] = 'x';
        }
        if ((constraint.structure[i] == '[') || (constraint.structure[i] == ']')){
            constraint.structure[i] = 'x';
        }
    }

    if (strlen(constraint.structure) != strlen(constraint.sequence)){
        fprintf(stderr, "\nSevere constraint error: sequence and structure have different lengths.\n\n");
        exit(14);
    }
    fclose(c_handle);
    fprintf(stdout, ".. done.\n");

    valid = checkStructure();

    if (valid == 0) {
        fprintf(stderr, "\nError: Your constraint is not a valid dot-bracket structure.\n");
        fprintf(stderr, "The number of opening brackets is unequal to the number of closing ones.\n\n");
        exit(15);
    }

    return valid;
}

/*--------------------------------------------------------------------------*/

int checkStructure() {

    int i, stack=0;

    for (i=0; i<strlen(constraint.structure); i++) {
        if (constraint.structure[i] == '(') stack++;
        else if (constraint.structure[i] == ')') stack--;
    }

    if (stack == 0) return 1;
    else return 0;

}

/*--------------------------------------------------------------------------*/

int identifyConstraint() {

    int i=0, idx=-1, j=0, k=0, k2=0;
    char *tmp = NULL;

    fprintf(stdout, "Identifying constraint '%s' in alignment ..\n", constraint.name);

    while (names[i] != NULL) {
        if (strcmp(constraint.name, names[i]) == 0) {
            tmp = (char*)space(strlen(seqs[i])+1*sizeof(char));
            for (j=0; j < strlen(seqs[i]); j++){
                if (isalpha(seqs[i][j])){
                    tmp[k] = seqs[i][j];
                    k++;
                }
            }
            if (strcmp(constraint.sequence, tmp) == 0){
                idx = i;
                break;
            }
            else 	free(tmp);
        }
        i++;
    }
    /* check identity of constraint name and seq. in both files. */

    if (idx == -1){
        fprintf(stderr, "\nError: Doubtless constraint identification within the alignment failed.\n");
        fprintf(stderr, "The sequence names (ID) in the constraint and alignment files differ,\n");
        fprintf(stderr, "or those sequences with the same ID differ in sequence. Aborting.\n\n");
        exit(16);
    }

    j=0; k=0;
    free(tmp);

    tmp = (char*)space(strlen(seqs[idx])+1*sizeof(char));
    constraint.expStructure = (char*)space((strlen(seqs[idx])+1)*sizeof(char));

    for (i=0; i<strlen(seqs[idx]); i++) {
        if (seqs[idx][i] != '-') {
            tmp[j] = seqs[idx][i];
            constraint.expStructure[k] = constraint.structure[k2];
            j++; k++; k2++;
        }
        else {
            constraint.expStructure[k] = '.';
            k++;
        }
    }
    tmp[j] = '\0';
    constraint.expStructure[k] = '\0';

    /* simple cross check if expansion ok */
    if (strcmp(constraint.sequence, tmp) != 0) {
        free(tmp);
        fprintf(stderr, "\nError: Doubtless constraint identification within the alignment failed.\n");
        fprintf(stderr, "The sequence names (ID) in the constraint and alignment files differ,\n");
        fprintf(stderr, "or those sequences with the same ID differ in sequence. Aborting.\n\n");
        exit(16);
    }
    else {
        fprintf(stdout, ".. done.\n");
        free(tmp);
        return idx;
    }

}

/*--------------------------------------------------------------------------*/

void weakenConstraint(int n, int c) {

    int l=0, i, j, k, co_i, stack = 0, retain = 0;
    char *col = NULL, *co_col = NULL;
    FILE *weak_constr_fp=NULL;

    fprintf(stdout, "Weakening constraint ..\n");

    i=1;
    while (seqs[i]){
        l = strlen(seqs[i]);
        if (l != strlen(seqs[i-1])){
            fprintf(stderr, "Severe error in alignment input file. Sequence lengths vary!\nAborting.\n\n");
            exit(88);
        }
        if (strlen(constraint.expStructure) != l) {
            fprintf(stderr, "Severe error in alignment input file. Sequence lengths vary!\nAborting.\n\n");
            exit(89);
        }
        i++;
    }

    constraint.weakened = (char*)space((strlen(constraint.expStructure)+1)*sizeof(char));
    strcpy(constraint.weakened, constraint.expStructure);

    for (i=0; i<l; i++) {
        if (constraint.weakened[i] == '.') {
            continue;
        }
        else if (constraint.weakened[i] == 'x') {
            continue;
        }
        else if (constraint.weakened[i] == '(') {
            col = (char*)space(n*sizeof(char));
            k = 0;
            for (j=0; j<n; j++) {
                if (j == c) continue;
                col[k] = seqs[j][i];
                k++;
            }
            col[k] = '\0';

            /* find corresponding position */
            co_i = i;
            stack = 1;
            while (stack) {
                co_i++;
                if (constraint.weakened[co_i] == '(') stack++;
                if (constraint.weakened[co_i] == ')') stack--;
            }

            co_col = (char*)space(n*sizeof(char));
            k = 0;
            for (j=0; j<n; j++) {
                if (j==c) continue;
                co_col[k] = seqs[j][co_i];
                k++;
            }
            co_col[k] = '\0';

            retain = scoreColPair(col, co_col, k);

            free(col);
            free(co_col);

            /* substitute if neccessary */
            if (retain == 0) {
                constraint.weakened[i] = '.';
                constraint.weakened[co_i] = '.';
            }
        }
        else if (constraint.weakened[i] == ')') {

            col = (char*)space(n*sizeof(char));
            k = 0;
            for (j=0; j<n; j++) {
                if (j == c) continue;
                col[k] = seqs[j][i];
                k++;
            }
            col[k] = '\0';

            /* find corresponding position */
            co_i = i;
            stack = 1;
            while (stack) {
                co_i--;
                if (constraint.weakened[co_i] == ')') stack++;
                if (constraint.weakened[co_i] == '(') stack--;
            }

            co_col = (char*)space(n*sizeof(char));
            k = 0;
            for (j=0; j<n; j++) {
                if (j == c) continue;
                co_col[k] = seqs[j][co_i];
                k++;
            }
            co_col[k] = '\0';

            retain = scoreColPair(col, co_col, k);

            free(col);
            free(co_col);

            /* substitute if neccessary */
            if (retain == 0) {
                constraint.weakened[i] = '.';
                constraint.weakened[co_i] = '.';
            }
        }
    }
    weak_constr_fp = fopen("SALSA_weakened_constr.txt", "w");
    fprintf(weak_constr_fp, "Structure constraint for the input alignment after weakening using stringency value switch -s1 set to %f:\n%s\n", options.SR1, constraint.weakened);
    fclose(weak_constr_fp);
    fprintf(stdout, ".. done.\n");
}

/*--------------------------------------------------------------------------*/

int scoreColPair(char *c1, char *c2, int n){

    int i, pairs = 0, noPairs = 0;
    float rel = 0.;

    int basepair[7][7] = {
                             /* _  A  C  G  U  R  N */
                             {0, 0, 0, 0, 0, 0, 0}, /* _ */
                             {0, 0, 0, 0, 1, 0, 0}, /* A */
                             {0, 0, 0, 1, 0, 0, 0}, /* C */
                             {0, 0, 1, 0, 1, 0, 0}, /* G */
                             {0, 1, 0, 1, 0, 1, 0}, /* U */
                             {0, 0, 0, 0, 1, 0, 0}, /* R */
                             {0, 0, 0, 0, 0, 0, 0}, /* N */
                         };
    int x, y;

    for (i=0; i<n; i++) {

        if (c1[i] == '_') x = 0;
        else if (c1[i] == 'A') x = 1;
        else if (c1[i] == 'C') x = 2;
        else if (c1[i] == 'G') x = 3;
        else if (c1[i] == 'U') x = 4;
        else if (c1[i] == 'R') x = 5;
        else x=6;

        if (c2[i] == '_') y = 0;
        else if (c2[i] == 'A') y = 1;
        else if (c2[i] == 'C') y = 2;
        else if (c2[i] == 'G') y = 3;
        else if (c2[i] == 'U') y = 4;
        else if (c2[i] == 'R') y = 5;
        else y=6;


        if (basepair[x][y] > 0) pairs++;
        else noPairs++;

    }

    if (pairs + noPairs > 0) {
        rel = (float) pairs/(float)(noPairs+pairs);
        if (rel >= options.SR1) return 1;
        else return 0;
    }
    else return 0;
}

/*--------------------------------------------------------------------------*/

void printConstraint() {

    fprintf(stdout, "Constraint:\n");
    fprintf(stdout, "%s\n",constraint.name);
    fprintf(stdout, "%s\n",constraint.sequence);
    fprintf(stdout, "%s\n",constraint.structure);
    fprintf(stdout, "%s\n",constraint.expStructure);
    fprintf(stdout, "%s\n",constraint.weakened);

}

/*--------------------------------------------------------------------------*/

void freeConstraint() {

    free(constraint.name);
    free(constraint.sequence);
    free(constraint.structure);
    free(constraint.expStructure);
    free(constraint.weakened);

}

