#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "utils.h"
#include "structs.h"

extern struct input_options options;
extern struct seqStruct *seqLst;
extern struct ALNdata *aln_final;
extern int all_possible_bp;
extern int final_conflict_warning;

char *consensus(int c, int r);
int **remove_final_conflicts(int **basepairLst, int **sperrLst);
void severe_error(int a);

/* ---------------------------------------- */

char *consensus(int c, int r){

  int a=0, aa=0, b=0, x, y;
  char *consensus=NULL, **all_folds=NULL;
  int i, ii, m, z, nix=0, sperr=0, sup_i=0, sup_ii=0, sup_value, d;
  int **seq_marker = NULL, **basepairLst = NULL, **sperrLst = NULL;
  int bracket, locked, conflicts, possible_bp;
  float float_sup_value, float_sperr;

  fprintf(stdout, "Calculating final consensus structure respecting set tolerance ..\n");

  consensus = (char*)space((r+1)*sizeof(char));
  for(a=0; a< r; a++){
    consensus[a] = '.';
  }
  consensus[r] = '\0';
  
  a=0;
  while(aln_final[c].seqname[a]) a++;
  all_folds = (char**)space((a+1)*sizeof(char*));
  all_folds[a] = NULL;

  a=0;
	while(aln_final[c].seqname[a]){
		while(strcmp(seqLst[b].name, aln_final[c].seqname[a])){
			b++;
		}
		
		seqLst[b].structure_aln = NULL;
		seqLst[b].structure_aln = (char*)space((r+1)*sizeof(char));
		y=0;
    
		for(x=0; x < r; x++){
			if(aln_final[c].alignment[a][x].nucl != '-'){
				seqLst[b].structure_aln[x] = seqLst[b].structure[y];
				y++;
			}
			else  seqLst[b].structure_aln[x] = '.';
		}
		seqLst[b].structure_aln[r] = '\0';

		all_folds[a] = NULL;
		all_folds[a] = (char*)space((r+1)*sizeof(char));
		strcpy(all_folds[a], seqLst[b].structure_aln);
		free(seqLst[b].structure_aln);
		b=0;
		a++;
	}	
	
	seq_marker = (int**)space(r*sizeof(int*));
	basepairLst = (int**)space((r*r)*sizeof(int*));
	sperrLst = (int**)space(r*sizeof(int*));
	for(i=0; i < r*r; i++){
		basepairLst[i] = NULL;
		basepairLst[i] = (int*)space(3*sizeof(int));
		basepairLst[i][0] = -1;
		basepairLst[i][1] = -1;
		basepairLst[i][2] = -1;
	}
    
	for(i=0; i< r; i++){
		seq_marker[i] = NULL;
		sperrLst[i] = NULL;
		seq_marker[i] = (int*)space(r*sizeof(int));
		sperrLst[i] = (int*)space(2*sizeof(int));
		sperrLst[i][0] = -1;
		sperrLst[i][1] = -1;
		for(ii=0; ii< r; ii++){
			seq_marker[i][ii] = 0;
		}
	}
    
	d=0;
	for(i=0; i< r; i++){
		z=0; sperr =0; nix =0;
		while(all_folds[z]){
			if(all_folds[z][i] == '.'){
				nix++;
			}
			if(all_folds[z][i] == 'x'){
				sperr++;
			}
	
			if(all_folds[z][i] == '('){
				m = 1;
				ii = i;
				while(m){
					ii++;
					if(all_folds[z][ii] == '(') m++;
					if(all_folds[z][ii] == ')') m--;
				}
				seq_marker[i][ii]++;
			}
			z++;
		}
    
		if(sperr){
			float_sperr = sperr / (float)a;
			if(float_sperr >= options.SR3){
				sperrLst[d][0] = i;
				sperrLst[d][1] = sperr;
				d++;
			}
		}
	}
  
	sup_value = 0; sup_i =-1; sup_ii =-1;
	for(i=0; i< r; i++){
		for(ii=0; ii< r; ii++){
			if(seq_marker[i][ii] > sup_value){
			sup_i = i;
			sup_ii = ii;
			sup_value = seq_marker[i][ii];
			}
		}
	}
  
	aa=0;
       
	float_sup_value = (sup_value / (float)a);
    
	while((float_sup_value > 0) && (float_sup_value >= options.SR3)){
		seq_marker[sup_i][sup_ii] = 0;
		basepairLst[aa][0] = sup_i;
		basepairLst[aa][1] = sup_ii;
		basepairLst[aa][2] = sup_value;
		sup_value = 0;
		aa++;
		for(i=0; i< r; i++){
			for(ii=0; ii< r; ii++){
				if(seq_marker[i][ii] > sup_value){
					sup_i = i;
					sup_ii = ii;
					sup_value = seq_marker[i][ii];
				}
			}
		}
		float_sup_value = (sup_value / (float)a);
	}
  
	basepairLst = remove_final_conflicts(basepairLst, sperrLst);

	a=0;
	while(sperrLst[a][0] != -1){
		if(sperrLst[a][1] != -1){
			consensus[sperrLst[a][0]] = 'x';
		}
		a++;
	}	
  
	a=0;
	while(basepairLst[a][0] != -1){
		if(basepairLst[a][2] != -1){
			if(basepairLst[a][0] < basepairLst[a][1]){
				consensus[basepairLst[a][0]] = '(';
				consensus[basepairLst[a][1]] = ')';
			}
			else severe_error(78);
		}
		a++;
	}

	for(i=0; i< r; i++){
		free(seq_marker[i]);
	}
	free(seq_marker);
  
	for(a=0; a< r; a++){
		free(sperrLst[a]);
	} 
	conflicts = 0; possible_bp = 0;
	for(a=0; a< (r * r); a++){
		if(basepairLst[a][0] != -1){
			all_possible_bp++;
			possible_bp++;
			if(basepairLst[a][2] == -1) conflicts++;
		}
		free(basepairLst[a]);
	}
	free(sperrLst);
	free(basepairLst);
  
	a=0; bracket = 0; locked = 0;
	while(consensus[a]){
		if(consensus[a] == '(') bracket++;
		if(consensus[a] == 'x') locked++;
		a++;
	}
  
	if(!possible_bp) possible_bp++;
	fprintf(stdout, "  . done.\nConsensus structure: %i pairs (%2.2f %% coverage of complete alignment length),\n                     %i conflict(s) have been resolved forcibly (%2.2f %%).\n", bracket, (float)bracket*100*2/(float)r, conflicts, conflicts*100/(float)possible_bp);
  
	fprintf(stdout, ".. finished.\n");
	return consensus;
}


/* ----------------------------------------------------- */

int **remove_final_conflicts(int **basepairLst, int **sperrLst){
  
  int a, b, conflict_count=1, q1, q2, w1, w2;
  
  a=0; b=0;
  while(conflict_count){
    conflict_count = 0;
    while(basepairLst[a][0] != -1){
      if(basepairLst[a][2] != -1){
	while(sperrLst[b][0] != -1){
	  if((sperrLst[b][0] == basepairLst[a][0]) || (sperrLst[b][0] == basepairLst[a][1])){
	    if(sperrLst[b][1] > basepairLst[a][2]){
	      basepairLst[a][2] = -1;
	      break;
	    }
	    if(sperrLst[b][1] < basepairLst[a][2]){
	      sperrLst[b][1] = -1;
	    }
	    if(sperrLst[b][1] == basepairLst[a][2]){
	      basepairLst[a][2] = -1;
	      conflict_count++;  /* not resolvable */
	      final_conflict_warning++;
	    }
	  }
	  b++;
	}
      }
      b=0;
      a++;
    }
  }
  
  a=0; b=1;
  if(!conflict_count) conflict_count =1;
  while(conflict_count){
    conflict_count = 0;
    while(basepairLst[a][0] != -1){
      
      if(basepairLst[a][2] != -1){
	while(basepairLst[b][0] != -1){
	  if(basepairLst[b][2] != -1){
	    q1 = basepairLst[a][0];
	    q2 = basepairLst[a][1];
	    w1 = basepairLst[b][0];
	    w2 = basepairLst[b][1];
	    if(((q1 <= w1) && (q2 >= w1) && (w2 >= q2)) || ((w1 <= q1) && (w2 >= q1) && (q2 >= w2))){
	      if(basepairLst[a][2] > basepairLst[b][2]){
		basepairLst[b][2] = -1;
	      }
	      if(basepairLst[a][2] < basepairLst[b][2]){
		basepairLst[a][2] = -1;  /* das kann ganz und gar nie passieren... */
		break;
	      }
	      if(basepairLst[a][2] == basepairLst[b][2]){
		basepairLst[b][2] = -1;
		conflict_count++;  /* not resolvable */
		final_conflict_warning++;
	      }
	      /* iterative optimization? */
	    }
	  }
	  b++;
	}
      }
      a++;
      b = a+1;
    }
  }

  return(basepairLst);
}


