#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "fold_vars.h"
#include "fold.h"
#include "utils.h"
#include "structs.h"
#include "scoretable.h"

struct input_options options;
struct seqStruct *seqLst;
struct conStruct constraint;
int alifold_count;
int conflict_warning;
int final_conflict_warning;
int all_possible_bp;
int number_of_seqs;
extern int overruling_scores[5];
int align_only;
int GAP_OPEN;
int GAP_EXT;
int NS_MATCH;
int NS_MISM;

void parseArgs(int c, char **argv);
void verify_scoretable();
int read_align_only(void);
void alignPairwise(int n, int c);
int readConstraint();
int readAln(void);
void encode(int c_idx);
int identifyConstraint();
void initStructs(int n);
void initStructs_align_only(int n);
void weakenConstraint(int n, int c);
int merge_all_structures(int seqno, int c);
int check_pairing(char c1, char c2);
int **remove_conflicts(int **basepairLst, int **sperrLst);
void write_indivConstr(int n, int **basepairLst, int **sperrLst);
void foldSingleSeqs(int n, int c);
void structaln(int n, int c_idx);
void freeStructs(int n);
void free_aln_arrays();
void freeConstraint();
void severe_error(int a);

/* ----------------------------------------------------- */

int main(int argc, char **argv) {
 
  int n_seq, c_idx, c_valid = 0, aa=0;
  
  options.SR1 = 0.6;
  options.SR2 = 0.6;
  options.SR3 = 0.6;
  options.plot = 1;
  options.isFasta = 0;
  options.kurzschluss = 0;
  alifold_count = 0;
  conflict_warning =0;
  final_conflict_warning =0;
  all_possible_bp = 0;
    
  fold_constrained = 1;
  dangles = 2;
  update_fold_params ();

  parseArgs(argc, argv);
    
  verify_scoretable();
  
	if(align_only) n_seq = read_align_only();
	else n_seq = readAln();    /* number of ALL sequences (incl. event. constraint) */
		
	seqLst = (struct seqStruct*)space((n_seq+1) * sizeof(struct seqStruct));
	number_of_seqs = n_seq; /* number of ALL sequences (incl. constraint) */
  
	if(align_only)  initStructs_align_only(n_seq);
	else {
		c_valid = readConstraint();
  
		c_idx = identifyConstraint(); /* index of constraint seq. */
  
		weakenConstraint(n_seq, c_idx);
  
		initStructs(n_seq);
  
		encode(c_idx);
  
		if(options.kurzschluss)  fprintf(stdout, "Retrieving all pairwise alignments from input ..\n");
		fprintf(stdout, "Folding all possible pairwise alignments ..\n");
  
		alignPairwise(n_seq, c_idx);
    
		aa = merge_all_structures(n_seq, c_idx);
  
		foldSingleSeqs(n_seq, c_idx);
	}
	
	structaln(n_seq, c_idx);
  
  fprintf(stdout, "\nThe files containing the results are named as follows:\n");
  fprintf(stdout, "\n  SALSA_structaln_sequ.aln (final structure guided multiple alignment)\n");
  fprintf(stdout, "  SALSA_structaln_struct.aln (final alignment - structure strings)\n");
  fprintf(stdout, "  SALSA_structaln_comb_typeA.fas (combined representation)\n");
  fprintf(stdout, "  SALSA_structaln_comb_typeB.fas (combined representation - ID added)\n");
  fprintf(stdout, "  SALSA_guide_tree.txt (multiple alignment guide tree)\n");
  if(!align_only) fprintf(stdout, "  SALSA_used_constr.txt (all used individual constraints)\n");
  if(!align_only) fprintf(stdout, "  SALSA_fold_results.txt (sequences and individual structures)\n");
  fprintf(stdout, "  SALSA_consensus.txt and SALSA_consensus_ss.ps (consensus structure)\n");
  fprintf(stdout, "\n  STRUC_*_ss.ps and CONS_*_ss.ps\n  (graphics of individual folding results and used constraints.)\n");
  if(aa) fprintf(stdout, "\n**WARNING**: There was not any structure information left for %i sequence(s)\nafter merging all pairwise alignment folds into one individual constraint.\nTry a less stringent setting for switch '-s2'.\n", aa);
   
  if(conflict_warning) fprintf(stdout, "\n**WARNING**: While merging pairwise alignment folds into one individual\nI could not resolve %i base pair conflicts in a doubtless way.\nThis is %2.2f %% of a total of %i possible pairs. Conflict removal was\nestimated and performed forcibly. Be careful...\n", conflict_warning, (conflict_warning * 100 / (float)all_possible_bp), all_possible_bp);

 if(final_conflict_warning) fprintf(stdout, "\n**WARNING**: While constructing final consensus structure I could\nnot resolve %i base pair conflicts in a doubtless way. Conflict removal\nwas estimated and performed forcibly. Be careful...\n", final_conflict_warning);

 if(align_only) fprintf(stdout, "\nRNAsalsa's combined sequence and structure alignment exited successfully.");
  fprintf(stdout, "\n%i RNA structures (alignments and individuals) have been calculated.\nThank you for using RNAsalsa. Good bye.\n\n", alifold_count);
  
  /* freeStructs(n_seq);
  freeConstraint();  */
  
  /*   free_aln_arrays(); */     
  /* double free..? - das muss ich noch klaeren. */
   
   /* free(seqLst); */
   return 0;
}

/* ----------------------------------------------------- */

int merge_all_structures(int seqno, int c){
  
  int n=0, i, ii, m, z, nix=0, sperr=0, sup_i=0, sup_ii=0, sup_value;
  int **seq_marker = NULL, **basepairLst = NULL, **sperrLst = NULL, a, aa=0, d=0;
  int bracket, locked, do_it, bb, conflicts, possible_bp;
  float float_sup_value, float_sperr, b=0.0;

  fprintf(stdout, "Calculating each individual consensus structure constraint ..\n");
  
  while(seqLst[n].name){
    if(n == c) {
      n++;
      continue;
    }

    fprintf(stdout, "  Optimization of individual constraint for %s ..\n", seqLst[n].name);
    
    b=0.0;
    bb=0;
    while(seqLst[n].consStrucLst[bb]){
      b = b + 1.0;
      bb++;
    }
    
    seq_marker = (int**)space(strlen(seqLst[n].sequence)*sizeof(int*));
    basepairLst = (int**)space((seqLst[n].length * seqLst[n].length)*sizeof(int*));
    sperrLst = (int**)space(strlen(seqLst[n].sequence)*sizeof(int*));
    for(i=0; i < (seqLst[n].length * seqLst[n].length); i++){
      basepairLst[i] = (int*)space(3*sizeof(int));
      basepairLst[i][0] = -1;
      basepairLst[i][1] = -1;
      basepairLst[i][2] = -1;
    }
    
    for(i=0; i<strlen(seqLst[n].sequence); i++){
      seq_marker[i] = (int*)space(strlen(seqLst[n].sequence)*sizeof(int));
      sperrLst[i] = (int*)space(2*sizeof(int));
      sperrLst[i][0] = -1;
      sperrLst[i][1] = -1;
      for(ii=0; ii<strlen(seqLst[n].sequence); ii++){
	seq_marker[i][ii] = 0;
      }
    }
    
    d=0;
    for(i=0; i<strlen(seqLst[n].sequence); i++){
      
      z=0; sperr =0; nix =0;
      
      while(seqLst[n].consStrucLst[z]){
	if(seqLst[n].consStrucLst[z][i] == '.'){
	  nix++;
	}
	if(seqLst[n].consStrucLst[z][i] == 'x'){
	  sperr++;
	}
	
	if(seqLst[n].consStrucLst[z][i] == '('){
	  m = 1;
	  ii = i;
	  while(m){
	    ii++;
	    if(seqLst[n].consStrucLst[z][ii] == '(') m++;
	    if(seqLst[n].consStrucLst[z][ii] == ')') m--;
	  }
	  
	  do_it = check_pairing(seqLst[n].sequence[i], seqLst[n].sequence[ii]);
	  if(do_it)   seq_marker[i][ii]++;
	}
	z++;
      }
      
      if(sperr){
	float_sperr = sperr / (float)b;
	if(float_sperr >= options.SR2){
	  sperrLst[d][0] = i;
	  sperrLst[d][1] = sperr;
	  d++;
	}
      }
    }
    
    sup_value = 0; sup_i =-1; sup_ii =-1;
    for(i=0; i<strlen(seqLst[n].sequence); i++){
      for(ii=0; ii<strlen(seqLst[n].sequence); ii++){
	if(seq_marker[i][ii] > sup_value){
	  sup_i = i;
	  sup_ii = ii;
	  sup_value = seq_marker[i][ii];
	}
      }
    }
    
    a=0;
       
    float_sup_value = (sup_value / (float)b);
    
    while((float_sup_value > 0) && (float_sup_value >= options.SR2)){
      seq_marker[sup_i][sup_ii] = 0;
      basepairLst[a][0] = sup_i;
      basepairLst[a][1] = sup_ii;
      basepairLst[a][2] = sup_value;
      sup_value = 0;
      a++;
      for(i=0; i<strlen(seqLst[n].sequence); i++){
	for(ii=0; ii<strlen(seqLst[n].sequence); ii++){
	  if(seq_marker[i][ii] > sup_value){
	    sup_i = i;
	    sup_ii = ii;
	    sup_value = seq_marker[i][ii];
	  }
	}
      }
      float_sup_value = (sup_value / (float)b);
    }
    
    if(!a)  aa++;
    
    basepairLst = remove_conflicts(basepairLst, sperrLst);
    
    write_indivConstr(n, basepairLst, sperrLst);
      
    for(i=0; i<strlen(seqLst[n].sequence); i++){
      free(seq_marker[i]);
    }
    free(seq_marker);
    
    for(a=0; a< strlen(seqLst[n].sequence); a++){
      free(sperrLst[a]);
    } 
    conflicts = 0; possible_bp = 0;
    for(a=0; a< (seqLst[n].length * seqLst[n].length); a++){
      if(basepairLst[a][0] != -1){
	all_possible_bp++;
	possible_bp++;
	if(basepairLst[a][2] == -1) conflicts++;
      }
      free(basepairLst[a]);
    }
    free(sperrLst);
    free(basepairLst);
    
    if(strlen(seqLst[n].indivConstr) != seqLst[n].length) severe_error(11);
    
    a=0; bracket = 0; locked = 0;
    while(seqLst[n].indivConstr[a]){
      if(seqLst[n].indivConstr[a] == '(') bracket++;
      if(seqLst[n].indivConstr[a] == 'x') locked++;
      a++;
    }
    
    if(!possible_bp) possible_bp++; /* Nur fuer die Ausgabe! */
    fprintf(stdout, "  . done. %i pairs (%2.2f %% coverage) with %i locked (%2.2f %%),\n          %i conflict(s) have been resolved forcibly (%2.2f %%).\n", bracket, (float)bracket*100*2/(float)seqLst[n].length, locked, (float)locked*100/(float)seqLst[n].length, conflicts, conflicts*100/(float)possible_bp);
    
    n++;
  }
  if(n != seqno) severe_error(110);
  fprintf(stdout, ". done.\n");
  return aa;
}

/* ----------------------------------------------------- */

int check_pairing(char c1, char c2){
    
  int basepair[7][7] = {
    /* _  A  C  G  U  R  N */
      {0, 0, 0, 0, 0, 0, 0}, /* _ */
      {0, 0, 0, 0, 1, 0, 0}, /* A */
      {0, 0, 0, 1, 0, 0, 0}, /* C */
      {0, 0, 1, 0, 1, 0, 0}, /* G */
      {0, 1, 0, 1, 0, 1, 0}, /* U */
      {0, 0, 0, 0, 1, 0, 0}, /* R */
      {0, 0, 0, 0, 0, 0, 0}, /* N */
  };
  
  int x, y;
  
  if(c1 == '_'){
    x = 0;
    severe_error(45);
  }
  else if(c1 == 'A') x = 1;
  else if(c1 == 'C') x = 2;
  else if(c1 == 'G') x = 3;
  else if(c1 == 'U') x = 4;
  else if(c1 == 'R') x = 5;
  else x=6;
	
  if(c2 == '_'){
    y = 0;
    severe_error(45);
  }
  else if(c2 == 'A') y = 1;
  else if(c2 == 'C') y = 2;
  else if(c2 == 'G') y = 3;
  else if(c2 == 'U') y = 4;
  else if(c2 == 'R') y = 5;
  else y=6;
	
  if(basepair[x][y] > 0) return 1;
  else return 0;
  
}

/* ----------------------------------------------------- */

int **remove_conflicts(int **basepairLst, int **sperrLst){
  
  int a, b, conflict_count=1, q1, q2, w1, w2;
  
  a=0; b=0;
  while(conflict_count){
    conflict_count = 0;
    while(basepairLst[a][0] != -1){
      if(basepairLst[a][2] != -1){
	while(sperrLst[b][0] != -1){
	  if((sperrLst[b][0] == basepairLst[a][0]) || (sperrLst[b][0] == basepairLst[a][1])){
	    if(sperrLst[b][1] > basepairLst[a][2]){
	      basepairLst[a][2] = -1;
	      break;
	    }
	    if(sperrLst[b][1] < basepairLst[a][2]){
	      sperrLst[b][1] = -1;
	    }
	    if(sperrLst[b][1] == basepairLst[a][2]){
	      basepairLst[a][2] = -1;
	      conflict_count++;  /* not resolvable */
	      conflict_warning++;
	    }
	  }
	  b++;
	}
      }
      b=0;
      a++;
    }
  }
  
  a=0; b=1;
  if(!conflict_count) conflict_count =1;
  while(conflict_count){
    conflict_count = 0;
    while(basepairLst[a][0] != -1){
      
      if(basepairLst[a][2] != -1){
	while(basepairLst[b][0] != -1){
	  if(basepairLst[b][2] != -1){
	    q1 = basepairLst[a][0];
	    q2 = basepairLst[a][1];
	    w1 = basepairLst[b][0];
	    w2 = basepairLst[b][1];
	    if(((q1 <= w1) && (q2 >= w1) && (w2 >= q2)) || ((w1 <= q1) && (w2 >= q1) && (q2 >= w2))){
	      if(basepairLst[a][2] > basepairLst[b][2]){
		basepairLst[b][2] = -1;
	      }
	      if(basepairLst[a][2] < basepairLst[b][2]){
		basepairLst[a][2] = -1;  /* das kann nie passieren... */
		break;
	      }
	      if(basepairLst[a][2] == basepairLst[b][2]){
		basepairLst[b][2] = -1;
		conflict_count++;  /* not resolvable */
		conflict_warning++;
	      }
	      /* iterative optimization? */
	    }
	  }
	  b++;
	}
      }
      a++;
      b = a+1;
    }
  }

  return(basepairLst);
}

/* ----------------------------------------------------- */

void write_indivConstr(int n, int **basepairLst, int **sperrLst){
  
  int a=0;
  
  while(sperrLst[a][0] != -1){
    if(sperrLst[a][1] != -1){
      seqLst[n].indivConstr[sperrLst[a][0]] = 'x';
    }
    a++;
  }
  
  a=0;
  while(basepairLst[a][0] != -1){
    if(basepairLst[a][2] != -1){
      if(basepairLst[a][0] < basepairLst[a][1]){
	seqLst[n].indivConstr[basepairLst[a][0]] = '(';
	seqLst[n].indivConstr[basepairLst[a][1]] = ')';
      }
      else severe_error(78);
    }
    a++;
  }
}
/* ----------------------------------------------------------------------- */

void verify_scoretable(void){

	if(overruling_scores[0]){
		GAP_OPEN = overruling_scores[1];
		GAP_EXT = overruling_scores[2];
		NS_MATCH = overruling_scores[3];
		NS_MISM = overruling_scores[4];
	}
	else {
		GAP_OPEN = STANDARD_GAP_OPEN;
		GAP_EXT = STANDARD_GAP_EXT;
		NS_MATCH = STANDARD_NS_MATCH;
		NS_MISM = STANDARD_NS_MISM;
	}
}

/* ----------------------------------------------------- */

void severe_error(int a){

  fprintf(stderr, "\nCRASH! A severe and unrecoverable error happened.\nVirtually, this never should have happened...\nRNAsalsa forced all processes to exit with error code %i.\nPlease send a bug report (including the error code) to roman@bioinf.uni-leipzig.de\nSORRY...\n\n", a);
  exit(a);
}
