#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "fold.h"
#include "utils.h"
#include "PS_dot.h"
#include "structs.h"

extern struct seqStruct *seqLst;
extern struct conStruct constraint;
extern struct input_options options;
extern char **seqs;
extern char **names;
extern char **strucs;
extern int fold_constrained;
extern int dangles;
extern int plot;
extern int alifold_count;

void initStructs(int n);
void initStructs_align_only(int n);
void printSingleInfo();
void foldSingleSeqs(int n, int c);
void freeStructs(int n);

/*--------------------------------------------------------------------------*/

void initStructs(int n) {
  
  int i, j=0, len, id=0;
  char *tmp_sequence = NULL;
  
  fprintf(stdout, "Initializing internal handling structs ..\n");
  
  while(names[id] != NULL) {
    
    len = strlen(names[id]);
    seqLst[id].name = NULL;
    seqLst[id].name = (char*)space((len+1)*sizeof(char));
    strcpy(seqLst[id].name, names[id]);
    
    len = strlen(seqs[id]);
    
    seqLst[id].alnSequence = NULL;
    seqLst[id].alnSequence = (char*)space((len+1)*sizeof(char));
    strcpy(seqLst[id].alnSequence, seqs[id]);
    strcat(seqLst[id].alnSequence, "\0");
    
    tmp_sequence = NULL;
    tmp_sequence = (char*)space((len+1)*sizeof(char));
    i=0;  j=0;
    while(seqs[id][j] != '\0') {
      if(seqs[id][j] != '-') {
	tmp_sequence[i] = seqs[id][j];
	i++;
      }
      j++;
    }
    tmp_sequence[i] = '\0';
    seqLst[id].sequence = NULL;
    seqLst[id].sequence = NULL;
    seqLst[id].sequence = (char*)space((strlen(tmp_sequence)+1)*sizeof(char));
    strcpy(seqLst[id].sequence, tmp_sequence);
    free(tmp_sequence);
    
    seqLst[id].structure = NULL;
    
    seqLst[id].length = strlen(seqLst[id].sequence);
    
    seqLst[id].mfe = 0.0;
    
    seqLst[id].consStrucLst = NULL;
    seqLst[id].consStrucLst = (char**)space((n-1)*sizeof(char*)); 
    i=0;
    while(i < n-1) { 
      seqLst[id].consStrucLst[i] = NULL;
      seqLst[id].consStrucLst[i] = (char*)space((seqLst[id].length+1)*sizeof(char));
      i++;
    }
    
    seqLst[id].consStrucLst[n-2] = NULL;
    
    seqLst[id].indivConstr = NULL;
    seqLst[id].indivConstr = (char*)space((seqLst[id].length+1)*sizeof(char));
    for(i=0; i< seqLst[id].length; i++){
      seqLst[id].indivConstr[i] = '.';
    }
    seqLst[id].indivConstr[seqLst[id].length] = '\0';
    id++;
  }
  
  seqLst[id].name = NULL;
  fprintf(stdout, ".. done.\n");
}

/*--------------------------------------------------------------------------*/

void initStructs_align_only(int n) {
  
  int len, id=0;
  
  fprintf(stdout, "Initializing internal handling structs ..\n");
  
  while(names[id] != NULL) {
    
    len = strlen(names[id]);
    seqLst[id].name = NULL;
    seqLst[id].name = (char*)space((len+1)*sizeof(char));
    strcpy(seqLst[id].name, names[id]);
    free(names[id]);
    
    len = strlen(seqs[id]);
    seqLst[id].sequence = NULL;
    seqLst[id].sequence = (char*)space((len+1)*sizeof(char));
    strcpy(seqLst[id].sequence, seqs[id]);
    strcat(seqLst[id].sequence, "\0");
    free(seqs[id]);
    
	seqLst[id].structure = NULL;
    seqLst[id].structure = (char*)space((len+1)*sizeof(char));
    strcpy(seqLst[id].structure, strucs[id]);
    strcat(seqLst[id].structure, "\0");
    free(strucs[id]);
	
	seqLst[id].length = len;
        
    id++;	
  }
  
   free(names);
	free(seqs);
	free(strucs);
  
  fprintf(stdout, ".. done.\n");
}

/*--------------------------------------------------------------------------*/

void printSingleInfo(){

  int i=0, j=0;
  
  while(seqLst[i].name){
    fprintf(stdout, "name: %s\n",seqLst[i].name);
    fprintf(stdout, "length: %d\n",seqLst[i].length);
    fprintf(stdout, "sequence: %s\n",seqLst[i].sequence);
    while(seqLst[i].consStrucLst[j]){
      fprintf(stdout, "constraint %d: %s\n", j, seqLst[i].consStrucLst[j]);
      j++;
    }
    j=0;
    fprintf(stdout, "individual constraint: %s\n",seqLst[i].indivConstr);
    fprintf(stdout, "fold -C result: %s\n",seqLst[i].structure);
    fprintf(stdout, "mfe: %5.2f\n",seqLst[i].mfe);
    fprintf(stdout, "\n");
    i++;
  }
}
/*--------------------------------------------------------------------------*/

void freeStructs(int n){
  
  int i, j;
  
  for(j=0; j<n; j++) {
    free(seqLst[j].name);
    free(seqLst[j].sequence);
    free(seqLst[j].structure);
    for(i=0; i<n-1; i++) {
      free(seqLst[j].consStrucLst[i]);
    }
    free(seqLst[j].consStrucLst);
    free(seqLst[j].alnSequence);
    free(seqLst[j].weakConstr);
    free(seqLst[j].weakConstr2);
    free(seqLst[j].indivConstr);
  }
}

/*--------------------------------------------------------------------------*/

void foldSingleSeqs(int n, int c) {

  int i, p=0, p_cons=0;
  FILE *fp = NULL;
  FILE *fp2 = NULL;
  char *filename = NULL;
  char *filename_cons = NULL;
  
  fprintf(stdout, "Folding sequences with individual constraints ..\n");
  
  fp= fopen("SALSA_fold_results.txt", "w");
  fp2= fopen("SALSA_used_constr.txt", "w");

  for(i=0; i<n; i++) {
    if(i == c) continue;

    seqLst[i].structure = (char*)space((seqLst[i].length+1)*sizeof(char));
    strcpy(seqLst[i].structure, seqLst[i].indivConstr);

    fprintf(fp2, ">%s\n%s\n%s\n", seqLst[i].name, seqLst[i].sequence, seqLst[i].structure); 
    fprintf(stdout, "  Constraint directed folding of %s ..\n", seqLst[i].name);

    seqLst[i].mfe = fold(seqLst[i].sequence, seqLst[i].structure);
    alifold_count++;

    fprintf(fp, ">%s\n%s\n%s\nmfe: %f\n", seqLst[i].name, seqLst[i].sequence, seqLst[i].structure, seqLst[i].mfe);
    
    if(options.plot == 1) { 
      filename = (char*)space((strlen(seqLst[i].name)+13)*sizeof(char));
      strcpy(filename, "STRUC_\0");
      strcat(filename, seqLst[i].name);
      strcat(filename, "_ss.ps\0");
      filename_cons = (char*)space((strlen(seqLst[i].name)+12)*sizeof(char));
      strcpy(filename_cons, "CONS_\0");
      strcat(filename_cons, seqLst[i].name);
      strcat(filename_cons, "_ss.ps\0");
	  p_cons = PS_rna_plot (seqLst[i].sequence, seqLst[i].indivConstr, filename_cons);
      p = PS_rna_plot (seqLst[i].sequence, seqLst[i].structure, filename);
      if(p==0)   fprintf(stderr, "PostScript file with structure graph could not be created.\n");
      if(p_cons==0)   fprintf(stderr, "PostScript file with constraint graph could not be created.\n");
      free(filename);
      free(filename_cons);
    }
    fprintf(stdout, "  .. done. Minimum free energy is %f\n", seqLst[i].mfe);
  }
  fclose(fp);
  fclose(fp2);
  printf(".. done.\n");
  
}
